/* RESTAURANT LUNCHES */
/* Results column */
div:has(.time-label-results-column) {
  background-color: powderblue;
  border-radius: 1.2rem;
  padding: 1rem;
  font-size: 1.8rem;
  font-weight: bold;
  line-height: 1.25rem;
  text-align: center;
}
/* Time Column */
div:has(.time-label-time-column) {
  background-color: powderblue;
  border-radius: .8rem;
  padding: .5rem 0rem;
  font-size: 1.25rem;
  font-weight: bold;
  line-height: 1.5rem;
  text-align: center;
}
/* TAKEAWAY LUNCHES */
/* Results column */
div:has(.takeaway-label-results-column) {
  background-color: #FFBA08;
  border-radius: 1.2rem;
  padding: 1rem;
  font-size: 1.8rem;
  font-weight: bold;
  line-height: 1.25rem;
  text-align: center;
}
/* Flexbox container for icon alignment */
.takeaway-label-results-column{
  display: flex;
  align-items: center;
  justify-content: flex-start;
}

/* Time Column */
div:has(.takeaway-label-time-column) {
  background-color: #FFBA08;
  border-radius: .8rem;
  padding: .5rem 0rem;
  font-size: 1.25rem;
  font-weight: bold;
  line-height: 1.5rem;
  text-align: center;
}

/* Flashing animation */
@keyframes fade-in {
  from {
    opacity: 0;
  }
  to {
    opacity: 1;
  }
}

.flashing-animation {
  animation: fade-in 1.5s ease infinite alternate;
}
