/* White background on primary color */
button.bk-active {
    background-color: white !important;
    box-shadow: none !important;
    color: var(--panel-primary-color);
}

.guest-override-flag {
    font-size: 1.25rem;
    color: var(--warning-bg-color);
    background-color: whitesmoke;
    border-left: 0.5rem solid var(--warning-bg-color);
    padding: 0.5rem
}

.icon-container{
    display: flex;
    align-items: center;
    justify-content: flex-start;
    gap: 0.5rem;
}

@keyframes fade-in {
    from {
      opacity: 0;
    }
    to {
      opacity: 1;
    }
}

.flashing-animation {
    animation: fade-in 1.5s ease infinite alternate;
}