<!-- Uses the template from https://github.com/bokeh/bokeh/tree/branch-3.2/examples/server/app/server_auth -->
<!DOCTYPE html>
<html>
<head>
    <title>Data-Lunch Logout</title>
    <link rel="icon" type="image/x-icon" href="/images/favicon.ico">
    <style>
        /* Import a special font for the title */
        @import url('https://fonts.googleapis.com/css2?family=Silkscreen:wght@700&display=swap');
        @import url('https://fonts.googleapis.com/css2?family=Space+Mono:wght@400&display=swap');
        /* Rules */
        *{
            margin:0;
            padding: 0;
            box-sizing: border-box;
        }
        html{
            height: 100%;
        }
        body{
            font-family: 'Segoe UI', sans-serif;;
            font-size: 1rem;
            line-height: 1.6;
            height: 100%;
        }
        p{
            padding-bottom: 5px;
        }
        .wrap {
            width: 100%;
            height: 100%;
            display: flex;
            justify-content: center;
            align-items: center;
            background: #fafafa;
        }
        .login-form{
            width: 350px;
            margin: 0 auto;
            border: 1px solid #ddd;
            padding: 2rem;
            background: #ffffff;
        }
        .form-input{
            background: #fafafa;
            border: 1px solid #eeeeee;
            padding: 12px;
            width: 100%;
        }
        .form-group{
            margin-bottom: 1rem;
            text-align: center;
        }
        .form-button{
            background: #0072B5;
            border: 1px solid #ddd;
            color: #ffffff;
            padding: 10px;
            width: 100%;
            text-transform: uppercase;
        }
        .form-button:hover{
            background: #005485;
        }
        .form-header{
            text-align: center;
            margin-bottom : 2rem;
        }
        .form-footer{
            font-family: "Space Mono", monospace;
            font-size: .8rem;
            text-align: center;
            margin-bottom : 2rem;
        }
        .form-header h1 {
        font-family: "Silkscreen", cursive;
        font-size: 2rem;
        line-height: 2rem;
        }
    </style>
</head>
<body>
    <div class="wrap">
        <form class="login-form" action=".{{ LOGIN_ENDPOINT }}" method="get">
            <div class="form-header">
                <h1>Data-Lunch</h1>
            </div>
            <div><small>
                <p class="form-footer"><i>Successfully logged out.</i></p>
            </small></div>
            <!--Login Button-->
            <div class="form-group">
                <button class="form-button" type="submit">Login</button>
            </div>
        </form>
    </div>
  </body>
  </html>
